#include <pthread.h>
#include <unistd.h>
#include <endian.h>
#include <linux/limits.h>

#include <sys/signal.h>

#include <mgba/core/core.h>
#include <mgba/core/serialize.h>

#include <msgba/packet/send_frame.h>
#include <msgba/packet.h>
#include <msgba/packet/hello.h>
#include <msgba/core_controller.h>

void
msPacketHelloDestroy(struct msPacketHello **hello) {
    if ((*hello)->rom) {
        free((*hello)->rom);
        (*hello)->rom = NULL;
    }
    if ((*hello)->savestate) {
        free((*hello)->savestate);
        (*hello)->savestate = NULL;
    }
    free(*hello);
    *hello = NULL;
}

void
msThreadCallbackStart(struct mCoreThread *threadContext) {
    sigset_t set;
    sigemptyset(&set);
    sigaddset(&set, SIGPIPE);
    pthread_sigmask(SIG_BLOCK, &set, NULL);
}

void
msThreadCallbackSetFrame(struct mCoreThread *threadContext) {
#define SEC_TO_NANO(sec) \
    (((uint64_t)sec) * 1000 * 1000 * 1000)
    struct msClientConnectionData *data = (struct msClientConnectionData *)threadContext->userData;
    color_t *outputBuffer;
    unsigned width, height;
    unsigned char *rawData = NULL;
    size_t rawDataLen = 0;
    struct msPacketSendFrame *sendFrame = NULL;
    struct msPacket *packet = NULL;
    struct timespec currentDate;
    timespec_get(&currentDate, TIME_UTC);
    uint64_t nanoCurrent = SEC_TO_NANO(currentDate.tv_sec % 1000) + currentDate.tv_nsec;
    uint64_t nanoLast = SEC_TO_NANO(data->lastFrameDate->tv_sec % 1000) + data->lastFrameDate->tv_nsec + (SEC_TO_NANO(1) / 60);

    if (!data || !data->coreController) {
        goto return_ms_thread_callback_set_frame;
    }
    outputBuffer = data->coreController->outputBuffer;

    while (nanoCurrent < nanoLast && nanoLast != 0 && nanoLast - nanoCurrent < SEC_TO_NANO(10)) {
	    usleep(500);
	    timespec_get(&currentDate, TIME_UTC);
	    nanoCurrent = SEC_TO_NANO(currentDate.tv_sec % 1000) + currentDate.tv_nsec;
    }
    *data->lastFrameDate = currentDate;

    data->coreController->threadContext.core->desiredVideoDimensions(data->coreController->threadContext.core, &width, &height);
    if (data->coreController->threadContext.core->platform(data->coreController->threadContext.core) == mPLATFORM_GB) {
        width = 160;
        height = 144;
    }

    if (data->numberFrame++ % 10 != 0) {
        return;
    }

    sendFrame = msPacketSendFrameNew(width, width * height, outputBuffer);
    rawData = msPacketSendFrameSerialize(sendFrame, &rawDataLen);
    if (!rawData) {
        goto return_ms_thread_callback_set_frame;
    }

    packet = msPacketNew(PACKET_SEND_FRAME, rawDataLen, rawData);
    msPacketSend(packet, data);

return_ms_thread_callback_set_frame:
    if (sendFrame) {
        msPacketSendFrameDestroy(&sendFrame);
    }
    if (packet) {
        msPacketDestroy(&packet);
    }
}

bool
msPacketHelloHandle(const struct msPacket *packet, struct msPacketHello *hello,
        struct msClientConnectionData *const data) {
    bool result = false;
    if (data->coreController) {
        fprintf(stderr, "Attempt to send a second hello, aborting connection.\n");
        goto return_ms_packet_hello_handle;
    }
    printf("Loading game and save for client_fd %d\n", data->clientFd);
    data->coreController = msCoreControllerLoadGame(hello->rom, hello->sizeRom, hello->savestate, hello->sizeSavestate, data);
    if (!data->coreController) {
        goto return_ms_packet_hello_handle;
    }
    msCoreControllerSetFrameCallback(data->coreController, &msThreadCallbackSetFrame);
    msCoreControllerSetStartCallback(data->coreController, &msThreadCallbackStart);
    msCoreControllerThreadStart(data->coreController);
    result = true;
return_ms_packet_hello_handle:
    msPacketHelloDestroy(&hello);
    return result;
}

bool
msPacketHelloGet(const struct msPacket *packet, int client_fd, 
        struct msClientConnectionData *const data) {
    bool result = false;
    struct msPacketHello *hello = NULL;
    unsigned char *rom = NULL;
    unsigned char *savestate = NULL;
    FILE *fp = fmemopen(packet->rawData, packet->size, "r");
    if (!fp) {
        printf("Unable to fmemopen\n");
        goto return_get_packet_hello;
    }
    size_t sizeRom = 0;
    size_t sizeSavestate = 0;
    size_t reads[4] = { 0,0,0,0 };
    size_t totalRead = 0;
    size_t to_read_size;

#define FREAD(target, size, count, fp, n, goto_to) \
    to_read_size  = size; \
    reads[n] = fread(target, to_read_size, count, fp); \
    reads[n] *= size; \
    if (reads[n] == -1) { \
        printf("Unable to read %s\n", #target); \
        goto goto_to; \
    }
    

    FREAD(&sizeRom, sizeof sizeRom, 1, fp, 0, return_get_packet_hello);
    sizeRom = be64toh(sizeRom);
    rom = malloc(sizeof *rom * sizeRom);

    FREAD(rom, sizeof *rom, sizeRom, fp, 1, return_get_packet_hello);

    FREAD(&sizeSavestate, sizeof sizeSavestate, 1, fp, 2, return_get_packet_hello);
    sizeSavestate = be64toh(sizeSavestate);
    savestate = malloc(sizeof *savestate * sizeSavestate);

    FREAD(savestate, sizeof *savestate, sizeSavestate, fp, 3, return_get_packet_hello);

    for (int i = 0; i<4; i++) {
        totalRead += reads[i];
    }
    if (totalRead != packet->size) {
        fprintf (stderr, "Total read (%lu) != packet->size (%lu)\n", totalRead, packet->size);
        goto return_get_packet_hello;
    }
    hello = calloc(1, sizeof *hello);
    hello->sizeRom = sizeRom;
    hello->rom = rom;
    hello->sizeSavestate = sizeSavestate;
    hello->savestate = savestate;
    result = true;
return_get_packet_hello:
    fclose(fp);
    if (!result) {
        fprintf(stderr, "Unable to read the packet hello\n");
        if (rom) {
            free (rom);
            rom = NULL;
        }
        if (savestate) {
            free (savestate);
            savestate = NULL;
        }
        return result;
    }
    return msPacketHelloHandle(packet, hello, data);
}

